import functools

from django.core.exceptions import FieldDoesNotExist
from django.db.models import ForeignKey, Prefetch
from django.db.models.constants import LOOKUP_SEP
from graphene.types.resolver import attr_resolver
from graphene_django import DjangoObjectType
from graphene_django.fields import DjangoListField
from graphql import ResolveInfo
from graphql.execution.base import (
    get_field_def,
)
from graphql.language.ast import FragmentSpread

from .utils import is_iterable


def query(queryset, info):
    return QueryOptimizer(info).optimize(queryset)


class QueryOptimizer(object):
    def __init__(self, info):
        self.root_info = info

    def optimize(self, queryset):
        info = self.root_info
        field_def = get_field_def(info.schema, info.parent_type, info.field_name)
        store = self._optimize_gql_selections(
            self._get_type(field_def),
            info.field_asts[0],
            info.parent_type,
        )
        return store.optimize_queryset(queryset)

    def _get_type(self, field_def):
        a_type = field_def.type
        if hasattr(a_type, 'of_type'):
            a_type = a_type.of_type
        return a_type

    def _optimize_gql_selections(self, field_type, field_ast, parent_type):
        store = QueryOptimizerStore()
        selection_set = field_ast.selection_set
        if not selection_set:
            return store
        model = field_type.graphene_type._meta.model
        for selection in selection_set.selections:
            name = selection.name.value
            if isinstance(selection, FragmentSpread):
                fragment = self.root_info.fragments[name]
                fragment_store = self._optimize_gql_selections(
                    field_type,
                    fragment,
                    parent_type,
                )
                store.append(fragment_store)
            else:
                selection_field_def = field_type.fields[name]
                self._optimize_field(
                    store,
                    model,
                    selection,
                    selection_field_def,
                    field_type,
                )
        return store

    def _optimize_field(self, store, model, selection, field_def, parent_type):
        optimized = self._optimize_field_by_name(store, model, selection, field_def, parent_type)
        optimized = self._optimize_field_by_hints(store, selection, field_def, parent_type) or optimized
        if not optimized:
            store.abort_only_optimization()

    def _optimize_field_by_name(self, store, model, selection, field_def, parent_type):
        name = self._get_name_from_resolver(field_def.resolver)
        if not name:
            return False
        model_field = self._get_model_field_from_name(model, name)
        if not model_field:
            return False
        if self._is_foreign_key_id(model_field, name):
            store.only(name)
            return True
        if model_field.many_to_one or model_field.one_to_one:
            field_store = self._optimize_gql_selections(
                self._get_type(field_def),
                selection,
                parent_type,
            )
            store.select_related(name, field_store)
            return True
        if model_field.one_to_many or model_field.many_to_many:
            field_store = self._optimize_gql_selections(
                self._get_type(field_def),
                selection,
                parent_type,
            )
            related_queryset = model_field.related_model.objects.all()
            store.prefetch_related(name, field_store, related_queryset)
            return True
        if not model_field.is_relation:
            store.only(name)
            return True
        return False

    def _get_optimization_hints(self, resolver):
        return getattr(resolver, 'optimization_hints', None)

    def _optimize_field_by_hints(self, store, selection, field_def, parent_type):
        optimization_hints = self._get_optimization_hints(field_def.resolver)
        if not optimization_hints:
            return False
        info = self._create_resolve_info(
            selection.name.value,
            (selection,),
            self._get_type(field_def),
            parent_type,
        )
        args = tuple(arg.value.value for arg in selection.arguments)
        self._add_optimization_hints(
            optimization_hints.select_related(info, *args),
            store.select_list,
        )
        self._add_optimization_hints(
            optimization_hints.prefetch_related(info, *args),
            store.prefetch_list,
        )
        if store.only_list is not None:
            self._add_optimization_hints(
                optimization_hints.only(info, *args),
                store.only_list,
            )
        return True

    def _add_optimization_hints(self, source, target):
        if source:
            if not is_iterable(source):
                source = (source,)
            target += source

    def _get_name_from_resolver(self, resolver):
        optimization_hints = self._get_optimization_hints(resolver)
        if optimization_hints:
            name = optimization_hints.model_field
            if name:
                return name
        if resolver == DjangoObjectType.resolve_id:
            return 'id'
        elif isinstance(resolver, functools.partial):
            resolver_fn = resolver
            if resolver_fn.func == DjangoListField.list_resolver:
                resolver_fn = resolver_fn.args[0]
            if resolver_fn.func == attr_resolver:
                return resolver_fn.args[0]

    def _get_model_field_from_name(self, model, name):
        try:
            return model._meta.get_field(name)
        except FieldDoesNotExist:
            descriptor = model.__dict__[name]
            return getattr(descriptor, 'rel', None) \
                or getattr(descriptor, 'related', None)  # Django < 1.9

    def _is_foreign_key_id(self, model_field, name):
        return (
            isinstance(model_field, ForeignKey) and
            model_field.name != name and
            model_field.get_attname() == name
        )

    def _create_resolve_info(self, field_name, field_asts, return_type, parent_type):
        return ResolveInfo(
            field_name,
            field_asts,
            return_type,
            parent_type,
            schema=self.root_info.schema,
            fragments=self.root_info.fragments,
            root_value=self.root_info.root_value,
            operation=self.root_info.operation,
            variable_values=self.root_info.variable_values,
            context=self.root_info.context,
        )


class QueryOptimizerStore():
    def __init__(self):
        self.select_list = []
        self.prefetch_list = []
        self.only_list = []

    def select_related(self, name, store):
        if store.select_list:
            for select in store.select_list:
                self.select_list.append(name + LOOKUP_SEP + select)
        else:
            self.select_list.append(name)
        for prefetch in store.prefetch_list:
            if isinstance(prefetch, Prefetch):
                prefetch.add_prefix(name)
            else:
                prefetch = name + LOOKUP_SEP + prefetch
            self.prefetch_list.append(prefetch)
        if self.only_list is not None:
            if store.only_list is None:
                self.abort_only_optimization()
            else:
                for only in store.only_list:
                    self.only_list.append(name + LOOKUP_SEP + only)

    def prefetch_related(self, name, store, queryset):
        if store.select_list or store.only_list:
            queryset = store.optimize_queryset(queryset)
            self.prefetch_list.append(Prefetch(name, queryset=queryset))
        elif store.prefetch_list:
            for prefetch in store.prefetch_list:
                self.prefetch_list.append(name + LOOKUP_SEP + prefetch)
        else:
            self.prefetch_list.append(name)

    def only(self, field):
        if self.only_list is not None:
            self.only_list.append(field)

    def abort_only_optimization(self):
        self.only_list = None

    def optimize_queryset(self, queryset):
        for select in self.select_list:
            queryset = queryset.select_related(select)
        for prefetch in self.prefetch_list:
            queryset = queryset.prefetch_related(prefetch)
        if self.only_list:
            queryset = queryset.only(*self.only_list)
        return queryset

    def append(self, store):
        self.select_list += store.select_list
        self.prefetch_list += store.prefetch_list
        if store.only_list is None:
            self.only_list = None
        else:
            self.only_list += store.only_list



from graphene.types.field import Field
from graphene.types.unmountedtype import UnmountedType

from .hints import OptimizationHints


def field(field_type, *args, **kwargs):
    if isinstance(field_type, UnmountedType):
        field_type = Field.mounted(field_type)

    optimization_hints = OptimizationHints(*args, **kwargs)
    get_resolver = field_type.get_resolver

    def get_optimized_resolver(parent_resolver):
        resolver = get_resolver(parent_resolver)
        resolver.optimization_hints = optimization_hints
        return resolver

    field_type.get_resolver = get_optimized_resolver
    return field_type



# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner >=4.0,<5dev'] if needs_pytest else []


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='graphene-django-optimizer',
    version='0.1.0',
    author='Tomás Fox',
    author_email='tomas.c.fox@gmail.com',
    description='Optimize database access inside graphene queries.',
    license='MIT',
    keywords='graphene django optimizer optimize graphql query prefetch select related',
    url='https://github.com/tfoxy/graphene-django-optimizer',
    packages=['graphene_django_optimizer'],
    setup_requires=pytest_runner,
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
    ],
)



from .utils import is_iterable, noop


def _normalize_hint_value(value):
    if not callable(value):
        if not is_iterable(value):
            value = (value,)
        return_value = value
        value = lambda *args, **kwargs: return_value
    return value


class OptimizationHints(object):
    def __init__(
        self,
        model_field=None,
        select_related=noop,
        prefetch_related=noop,
        only=noop,
    ):
        self.model_field = model_field
        self.prefetch_related = _normalize_hint_value(prefetch_related)
        self.select_related = _normalize_hint_value(select_related)
        self.only = _normalize_hint_value(only)
